html,
body {
	position: relative;
	width: 100%;
	height: 100%;
}

body {
	color: #333;
	margin: 0;
	padding: 0;
	box-sizing: border-box;
	font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto,
		Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif;
}

h2 {
	font-weight: 500;
}

.mdc-data-table__pagination {
	border-top-width: 0px;
}

table {
	overflow: visible;
}

.end-cell {
	border-right: 1px solid #e0e0e0;
}

.remove {
	cursor: pointer;
}
